"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ValidatorMain = void 0;
const path_1 = __importDefault(require("path"));
const defined_1 = require("./base/defined");
const readJsonUnchecked_1 = require("./base/readJsonUnchecked");
const globMatcher_1 = require("./base/globMatcher");
const writeUnchecked_1 = require("./base/writeUnchecked");
const Iterables_1 = require("./base/Iterables");
const Validators_1 = require("./validation/Validators");
const ExtendedObjectsValidators_1 = require("./validation/ExtendedObjectsValidators");
const BoundingVolumeS2Validator_1 = require("./validation/extensions/BoundingVolumeS2Validator");
/**
 * A class summarizing the command-line functions of the validator.
 *
 * The functions in this class are supposed to be called from `main.ts`,
 * based on the parsed command line arguments.
 *
 * @internal
 */
class ValidatorMain {
    static async validateTilesetFile(fileName, reportFileName) {
        console.log("Validating tileset " + fileName);
        const validationResult = await Validators_1.Validators.validateTilesetFile(fileName);
        if ((0, defined_1.defined)(reportFileName)) {
            await (0, writeUnchecked_1.writeUnchecked)(reportFileName, validationResult.serialize());
        }
        else {
            console.log("Validation result:");
            console.log(validationResult.serialize());
        }
        return validationResult;
    }
    static async validateTilesetsDirectory(directoryName, globPattern, writeReports) {
        console.log("Validating tilesets from " + directoryName + " matching " + globPattern);
        const allFiles = Iterables_1.Iterables.overFiles(directoryName);
        const ignoreCase = true;
        const matcher = (0, globMatcher_1.globMatcher)(globPattern, ignoreCase);
        const tilesetFiles = Iterables_1.Iterables.filter(allFiles, matcher);
        let numFiles = 0;
        let numFilesWithErrors = 0;
        let numFilesWithWarnings = 0;
        let numFilesWithInfos = 0;
        for (const tilesetFile of tilesetFiles) {
            let reportFileName = undefined;
            if (writeReports) {
                reportFileName = ValidatorMain.deriveReportFileName(tilesetFile);
            }
            const validationResult = await ValidatorMain.validateTilesetFile(tilesetFile, reportFileName);
            numFiles++;
            if (validationResult.numErrors > 0) {
                numFilesWithErrors++;
            }
            if (validationResult.numWarnings > 0) {
                numFilesWithWarnings++;
            }
            if (validationResult.numInfos > 0) {
                numFilesWithInfos++;
            }
        }
        console.log("Validated " + numFiles + " files");
        console.log("    " + numFilesWithErrors + " files with errors");
        console.log("    " + numFilesWithWarnings + " files with warnings");
        console.log("    " + numFilesWithInfos + " files with infos");
    }
    static async validateSchemaFile(fileName, reportFileName) {
        console.log("Validating schema " + fileName);
        const validationResult = await Validators_1.Validators.validateSchemaFile(fileName);
        if ((0, defined_1.defined)(reportFileName)) {
            await (0, writeUnchecked_1.writeUnchecked)(reportFileName, validationResult.serialize());
        }
        else {
            console.log("Validation result:");
            console.log(validationResult.serialize());
        }
        return validationResult;
    }
    static async validateSubtreeFile(fileName, validationState, implicitTiling, reportFileName) {
        console.log("Validating subtree " + fileName);
        const validationResult = await Validators_1.Validators.validateSubtreeFile(fileName, validationState, implicitTiling);
        if ((0, defined_1.defined)(reportFileName)) {
            await (0, writeUnchecked_1.writeUnchecked)(reportFileName, validationResult.serialize());
        }
        else {
            console.log("Validation result:");
            console.log(validationResult.serialize());
        }
        return validationResult;
    }
    static async validateAllTilesetSpecFiles(writeReports) {
        const recurse = true;
        const allSpecFiles = Iterables_1.Iterables.overFiles(ValidatorMain.specsDataRootDir + "/tilesets", recurse);
        const ignoreCase = true;
        const matcher = (0, globMatcher_1.globMatcher)("**/*.json", ignoreCase);
        const specFiles = Iterables_1.Iterables.filter(allSpecFiles, matcher);
        for (const specFile of specFiles) {
            let reportFileName = undefined;
            if (writeReports) {
                reportFileName = ValidatorMain.deriveReportFileName(specFile);
            }
            await ValidatorMain.validateTilesetFile(specFile, reportFileName);
        }
    }
    static async validateAllMetadataSchemaSpecFiles(writeReports) {
        const recurse = false;
        const allSpecFiles = Iterables_1.Iterables.overFiles(ValidatorMain.specsDataRootDir + "schemas", recurse);
        const ignoreCase = true;
        const matcher = (0, globMatcher_1.globMatcher)("**/*.json", ignoreCase);
        const specFiles = Iterables_1.Iterables.filter(allSpecFiles, matcher);
        for (const specFile of specFiles) {
            let reportFileName = undefined;
            if (writeReports) {
                reportFileName = ValidatorMain.deriveReportFileName(specFile);
            }
            await ValidatorMain.validateSchemaFile(specFile, reportFileName);
        }
    }
    static async validateAllSubtreeSpecFiles(writeReports) {
        const recurse = false;
        const allSpecFiles = Iterables_1.Iterables.overFiles(ValidatorMain.specsDataRootDir + "subtrees", recurse);
        const ignoreCase = true;
        const matcher = (0, globMatcher_1.globMatcher)("**/{*.json,*.subtree}", ignoreCase);
        const specFiles = Iterables_1.Iterables.filter(allSpecFiles, matcher);
        for (const specFile of specFiles) {
            let reportFileName = undefined;
            if (writeReports) {
                reportFileName = ValidatorMain.deriveReportFileName(specFile);
            }
            await ValidatorMain.validateSubtreeSpecFile(specFile, reportFileName);
        }
    }
    static async validateSubtreeSpecFile(fileName, reportFileName) {
        let implicitTiling = undefined;
        let validationState = {
            hasSchemaDefinition: false,
            hasGroupsDefinition: false,
        };
        // The `TileImplicitTiling` object that defines the
        // structure of subtrees in the specs directory
        const specImplicitTiling = await (0, readJsonUnchecked_1.readJsonUnchecked)("specs/data/subtrees/validSubtreeImplicitTiling.json.input");
        implicitTiling = specImplicitTiling;
        // The `ValidationState` object that contains the
        // schema for the subtrees in the specs directory
        const specSchema = await (0, readJsonUnchecked_1.readJsonUnchecked)("specs/data/schemas/validSchema.json");
        const specValidationState = {
            hasSchemaDefinition: true,
            validatedSchema: specSchema,
            hasGroupsDefinition: false,
            validatedGroups: undefined,
        };
        validationState = specValidationState;
        await ValidatorMain.validateSubtreeFile(fileName, validationState, implicitTiling, reportFileName);
    }
    /**
     * Register the validators for known extensions
     */
    static registerExtensionValidators() {
        // Register the validator for 3DTILES_bounding_volume_S2
        {
            const s2Validator = new BoundingVolumeS2Validator_1.BoundingVolumeS2Validator();
            const override = true;
            ExtendedObjectsValidators_1.ExtendedObjectsValidators.register("3DTILES_bounding_volume_S2", s2Validator, override);
        }
        // Register an empty validator for 3DTILES_content_gltf
        // (The extension does not have any properties to be
        // validated)
        {
            const emptyValidator = Validators_1.Validators.createEmptyValidator();
            const override = false;
            ExtendedObjectsValidators_1.ExtendedObjectsValidators.register("3DTILES_content_gltf", emptyValidator, override);
        }
    }
    /**
     * Derives a file name for a report from the given input file name.
     * The resulting file name will be a file in the same directory as
     * the given one. Further details are intentionally not specified here.
     *
     * @param inputFileName - The input file name
     * @returns The report file name
     */
    static deriveReportFileName(inputFileName) {
        const basename = path_1.default.basename(inputFileName, path_1.default.extname(inputFileName));
        const extension = ".report.json";
        return path_1.default.join(path_1.default.dirname(inputFileName), basename + extension);
    }
}
exports.ValidatorMain = ValidatorMain;
ValidatorMain.specsDataRootDir = "specs/data/";
