"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ArchiveFunctions3tz = void 0;
const fs_1 = __importDefault(require("fs"));
const crypto_1 = __importDefault(require("crypto"));
const defined_1 = require("./base/defined");
const TilesetArchiveError_1 = require("./TilesetArchiveError");
class ArchiveFunctions3tz {
    static getLastCentralDirectoryEntry(fd, stat) {
        const bytesToRead = 320;
        const buffer = Buffer.alloc(bytesToRead);
        const offset = stat.size - bytesToRead;
        const length = bytesToRead;
        fs_1.default.readSync(fd, buffer, 0, length, offset);
        let start = 0, end = 0;
        for (let i = buffer.length - 4; i > 0; i--) {
            const val = buffer.readUInt32LE(i);
            if (val === ArchiveFunctions3tz.ZIP_END_OF_CENTRAL_DIRECTORY_HEADER_SIG) {
                end = i;
            }
            if (val === ArchiveFunctions3tz.ZIP_START_OF_CENTRAL_DIRECTORY_HEADER_SIG) {
                start = i;
                break;
            }
        }
        if (start !== end) {
            return buffer.subarray(start);
        }
        return undefined;
    }
    static getFileContents(fd, buffer, expectedFilename) {
        const comp_size = buffer.readUInt32LE(20);
        const filename_size = buffer.readUInt16LE(28);
        const extra_size = buffer.readUInt16LE(30);
        const filename = buffer.toString("utf8", ArchiveFunctions3tz.ZIP_CENTRAL_DIRECTORY_STATIC_SIZE, ArchiveFunctions3tz.ZIP_CENTRAL_DIRECTORY_STATIC_SIZE + filename_size);
        if (filename !== expectedFilename) {
            throw new TilesetArchiveError_1.TilesetArchiveError(`Central Directory File Header filename was ${filename}, expected ${expectedFilename}`);
        }
        let offset = buffer.readUInt32LE(42);
        // if we get this offset, then the offset is stored in the 64 bit extra field
        if (offset === 0xffffffff) {
            let offset64Found = false;
            const endExtrasOffset = ArchiveFunctions3tz.ZIP_CENTRAL_DIRECTORY_STATIC_SIZE +
                filename_size +
                extra_size;
            let currentOffset = ArchiveFunctions3tz.ZIP_CENTRAL_DIRECTORY_STATIC_SIZE + filename_size;
            while (!offset64Found && currentOffset < endExtrasOffset) {
                const extra_tag = buffer.readUInt16LE(currentOffset);
                const extra_size = buffer.readUInt16LE(currentOffset + 2);
                if (extra_tag ===
                    ArchiveFunctions3tz.ZIP64_EXTENDED_INFORMATION_EXTRA_SIG &&
                    extra_size == 8) {
                    offset = Number(buffer.readBigUInt64LE(currentOffset + 4));
                    offset64Found = true;
                }
                else {
                    currentOffset += extra_size;
                }
            }
            if (!offset64Found) {
                throw new TilesetArchiveError_1.TilesetArchiveError("No zip64 extended offset found");
            }
        }
        const localFileDataSize = ArchiveFunctions3tz.ZIP_LOCAL_FILE_HEADER_STATIC_SIZE +
            filename_size +
            +48 /* over-estimated local file header extra field size, to try and read all data in one go */ +
            comp_size;
        const localFileDataBuffer = Buffer.alloc(localFileDataSize);
        fs_1.default.readSync(fd, localFileDataBuffer, 0, localFileDataSize, offset);
        // ok, skip past the filename and extras and we have our data
        const local_comp_size = localFileDataBuffer.readUInt32LE(18);
        const local_filename_size = localFileDataBuffer.readUInt16LE(26);
        const local_extra_size = localFileDataBuffer.readUInt16LE(28);
        const dataStartOffset = ArchiveFunctions3tz.ZIP_LOCAL_FILE_HEADER_STATIC_SIZE +
            local_filename_size +
            local_extra_size;
        const fileDataBuffer = localFileDataBuffer.slice(dataStartOffset, dataStartOffset + local_comp_size);
        if (fileDataBuffer.length === 0) {
            throw new TilesetArchiveError_1.TilesetArchiveError(`Failed to get file data at offset ${dataStartOffset}`);
        }
        return fileDataBuffer;
    }
    static parseIndexData(buffer) {
        if (buffer.length % 24 !== 0) {
            throw new TilesetArchiveError_1.TilesetArchiveError(`Bad index buffer length: ${buffer.length}`);
        }
        const numEntries = buffer.length / 24;
        const index = [];
        //console.log(`Zip index contains ${numEntries} entries.`);
        for (let i = 0; i < numEntries; i++) {
            const byteOffset = i * 24;
            const hash = buffer.slice(byteOffset, byteOffset + 16);
            const offset = buffer.readBigUInt64LE(byteOffset + 16);
            index.push({ hash: hash, offset: offset });
        }
        return index;
    }
    static md5LessThan(md5hashA, md5hashB) {
        const aLo = md5hashA.readBigUInt64LE();
        const bLo = md5hashB.readBigUInt64LE();
        if (aLo === bLo) {
            const aHi = md5hashA.readBigUInt64LE(8);
            const bHi = md5hashB.readBigUInt64LE(8);
            return aHi < bHi;
        }
        return aLo < bLo;
    }
    static zipIndexFind(zipIndex, searchHash) {
        let low = 0;
        let high = zipIndex.length - 1;
        while (low <= high) {
            const mid = Math.floor(low + (high - low) / 2);
            const entry = zipIndex[mid];
            //console.log(`mid: ${mid} entry: ${entry.md5hash.toString('hex')}`);
            if (entry.hash.compare(searchHash) === 0) {
                return mid;
            }
            else if (ArchiveFunctions3tz.md5LessThan(zipIndex[mid].hash, searchHash)) {
                low = mid + 1;
            }
            else {
                high = mid - 1;
            }
        }
        return -1;
    }
    static searchIndex(zipIndex, searchPath) {
        const hashedSearchPath = crypto_1.default
            .createHash("md5")
            .update(searchPath)
            .digest();
        //console.log(`Searching index for ${searchPath} (${hashedSearchPath.toString('hex')})`);
        //console.time('Search index');
        const matchedIndex = ArchiveFunctions3tz.zipIndexFind(zipIndex, hashedSearchPath);
        //console.log(`matchedIndex: ${matchedIndex}`);
        //console.timeEnd('Search index');
        if (matchedIndex === -1) {
            console.log(`Couldn't find ${searchPath} (${hashedSearchPath.toString("hex")})`);
            return undefined;
        }
        const entry = zipIndex[matchedIndex];
        //console.log(`Matched index: ${matchedIndex} - offset: ${entry.offset}`);
        return entry;
    }
    static parseLocalFileHeader(buffer, expectedFilename) {
        const signature = buffer.readUInt32LE(0);
        if (signature !== 0x04034b50) {
            throw new TilesetArchiveError_1.TilesetArchiveError(`Bad local file header signature: 0x${signature.toString(16)}`);
        }
        const comp_size = buffer.readUInt32LE(18);
        const filename_size = buffer.readUInt16LE(26);
        const extra_size = buffer.readUInt16LE(28);
        const filename = buffer.toString("utf8", ArchiveFunctions3tz.ZIP_LOCAL_FILE_HEADER_STATIC_SIZE, ArchiveFunctions3tz.ZIP_LOCAL_FILE_HEADER_STATIC_SIZE + filename_size);
        if (filename !== expectedFilename) {
            throw new TilesetArchiveError_1.TilesetArchiveError(`Local File Header filename was ${filename}, expected ${expectedFilename}`);
        }
        const compressedSize = comp_size;
        if (compressedSize === 0) {
            throw new TilesetArchiveError_1.TilesetArchiveError("Zip Local File Headers must have non-zero file sizes set.");
        }
        return {
            signature: signature,
            comp_size: comp_size,
            filename_size: filename_size,
            extra_size: extra_size,
        };
    }
    static readZipLocalFileHeader(fd, offset, path) {
        const headerSize = ArchiveFunctions3tz.ZIP_LOCAL_FILE_HEADER_STATIC_SIZE + path.length;
        const headerBuffer = Buffer.alloc(headerSize);
        //console.log(`readZipLocalFileHeader path: ${path} headerSize: ${headerSize} offset: ${offset}`);
        fs_1.default.readSync(fd, headerBuffer, 0, headerSize, Number(offset));
        //console.log(`headerBuffer: ${result.buffer}`);
        const header = ArchiveFunctions3tz.parseLocalFileHeader(headerBuffer, path);
        //console.log(header);
        return header;
    }
    static normalizePath(path) {
        // on Windows, the paths get backslashes (due to path.join)
        // normalize that to be able to deal with internal zip paths
        const res = path.replace(/\.\//, "");
        return res.replace(/\\/g, "/");
    }
    static readZipIndex(fd) {
        const stat = fs_1.default.fstatSync(fd);
        const centralDirectoryEntryData = ArchiveFunctions3tz.getLastCentralDirectoryEntry(fd, stat);
        if (!(0, defined_1.defined)(centralDirectoryEntryData)) {
            throw new TilesetArchiveError_1.TilesetArchiveError("Could not read last central directory entry");
        }
        const indexFileName = "@3dtilesIndex1@";
        const indexFileContents = ArchiveFunctions3tz.getFileContents(fd, centralDirectoryEntryData, indexFileName);
        const zipIndex = ArchiveFunctions3tz.parseIndexData(indexFileContents);
        return zipIndex;
    }
    static readFileName(fd, offset) {
        const headerSize = ArchiveFunctions3tz.ZIP_LOCAL_FILE_HEADER_STATIC_SIZE + 320;
        const headerBuffer = Buffer.alloc(headerSize);
        fs_1.default.readSync(fd, headerBuffer, 0, headerSize, Number(offset));
        const filename_size = headerBuffer.readUInt16LE(26);
        const filename = headerBuffer.toString("utf8", ArchiveFunctions3tz.ZIP_LOCAL_FILE_HEADER_STATIC_SIZE, ArchiveFunctions3tz.ZIP_LOCAL_FILE_HEADER_STATIC_SIZE + filename_size);
        return filename;
    }
    static readEntryData(fd, zipIndex, path) {
        const normalizedPath = ArchiveFunctions3tz.normalizePath(path);
        const match = ArchiveFunctions3tz.searchIndex(zipIndex, normalizedPath);
        if ((0, defined_1.defined)(match)) {
            const header = ArchiveFunctions3tz.readZipLocalFileHeader(fd, match.offset, path);
            const fileDataOffset = Number(match.offset) +
                ArchiveFunctions3tz.ZIP_LOCAL_FILE_HEADER_STATIC_SIZE +
                header.filename_size +
                header.extra_size;
            const fileContentsBuffer = Buffer.alloc(header.comp_size);
            //console.log(`Fetching data at offset ${fileDataOffset} size: ${header.comp_size}`);
            fs_1.default.readSync(fd, fileContentsBuffer, 0, header.comp_size, fileDataOffset);
            return fileContentsBuffer;
        }
        //console.log('No entry found for path ', path)
        return undefined;
    }
}
exports.ArchiveFunctions3tz = ArchiveFunctions3tz;
ArchiveFunctions3tz.ZIP_END_OF_CENTRAL_DIRECTORY_HEADER_SIG = 0x06054b50;
ArchiveFunctions3tz.ZIP_START_OF_CENTRAL_DIRECTORY_HEADER_SIG = 0x02014b50;
ArchiveFunctions3tz.ZIP64_EXTENDED_INFORMATION_EXTRA_SIG = 0x0001;
ArchiveFunctions3tz.ZIP_LOCAL_FILE_HEADER_STATIC_SIZE = 30;
ArchiveFunctions3tz.ZIP_CENTRAL_DIRECTORY_STATIC_SIZE = 46;
