/**
 * Utility functions for generic operations on values that
 * may be numbers or (potentially multi-dimensional) arrays
 * of numbers.
 *
 * These methods are mainly used for performing operations
 * on metadata values that have been found to be numeric
 * values (i.e. SCALAR values, VECn or MATn values, or
 * arrays thereof)
 *
 * When two values are involved, then the methods assume
 * that the values have the same structure, i.e. they are
 * both numbers or arrays with the same length.
 */
export declare class ArrayValues {
    /**
     * Multiplies the given input value with the given factor.
     *
     * If the factor is undefined, then the original value
     * is returned.
     *
     * This considers the case that the values are numbers or
     * (potentially multi-dimensional) arrays of numbers.
     *
     * @param value - The input value
     * @param factor - The optional factor
     * @returns The resulting value
     */
    static deepMultiply(value: any, factor: any): any;
    /**
     * Adds the given addend to the given input value.
     *
     * If the addend is undefined, then the original value
     * is returned.
     *
     * This considers the case that the values are numbers or
     * (potentially multi-dimensional) arrays of numbers.
     *
     * @param value - The input value
     * @param addend - The optional addend
     * @returns The resulting value
     */
    static deepAdd(value: any, addend: any): any;
    /**
     * Computes the minimum of the given values.
     *
     * This considers the case that the values are numbers or
     * (potentially multi-dimensional) arrays of numbers,
     * and computes the component-wise minimum.
     *
     * @param a - The first value
     * @param b - THe second value
     * @returns The mimimum value
     */
    static deepMin(a: any, b: any): any;
    /**
     * Computes the maximum of the given values.
     *
     * This considers the case that the values are numbers or
     * (potentially multi-dimensional) arrays of numbers,
     * and computes the component-wise maximum.
     *
     * @param a - The first value
     * @param b - THe second value
     * @returns The maximum value
     */
    static deepMax(a: any, b: any): any;
    /**
     * Returns whether the given values are equal up to the
     * give relative epsilon.
     *
     * This considers the case that the values are numbers or
     * (potentially multi-dimensional) arrays of numbers.
     *
     * @param a - The first element
     * @param b - The second element
     * @param epsilon - A relative epsilon
     * @returns Whether the objects are equal
     */
    static deepEqualsEpsilon(a: any, b: any, epsilon: number): boolean;
    /**
     * From CesiumJS:
     *
     * Returns whether two numbers are equal, up to a certain epsilon
     *
     * @param left - The first value
     * @param right - The second value
     * @param relativeEpsilon - The maximum inclusive delta for the relative tolerance test.
     * @param absoluteEpsilon - The maximum inclusive delta for the absolute tolerance test.
     * @returns Whether the values are equal within the epsilon
     */
    private static equalsEpsilon;
    /**
     * Checks whether two values are equal.
     *
     * This considers the case that the values are numbers or
     * (potentially multi-dimensional) arrays of numbers.
     *
     * @param a - The first value
     * @param b - The second value
     * @returns Whether the values are equal
     */
    static deepEquals(a: any, b: any): boolean;
    /**
     * Returns a deep clone of the given value.
     *
     * This considers the case that the values are (potentially
     * multi-dimensional) arrays. Non-array values (including
     * objects!) will be returned directly.
     *
     * @param value - The input value
     * @returns The result value
     */
    static deepClone(value: any): any;
    /**
     * Returns whether one value is less than another.
     *
     * This considers the case that the values are numbers or
     * (potentially multi-dimensional) arrays of numbers.
     *
     * It returns whether the first number is smaller than
     * the second number. For arrays, it recursively checks
     * whether ANY element of the first array is smaller
     * than the corresponding element of the secon array.
     *
     * @param a - The first value
     * @param b - The second value
     * @returns Whether the first value is less than the second
     */
    static anyDeepLessThan(a: any, b: any): boolean;
    /**
     * Returns whether one value is greater than another.
     *
     * This considers the case that the values are numbers or
     * (potentially multi-dimensional) arrays of numbers.
     *
     * It returns whether the first number is greater than
     * the second number. For arrays, it recursively checks
     * whether ANY element of the first array is greater
     * than the corresponding element of the secon array.
     *
     * @param a - The first value
     * @param b - The second value
     * @returns Whether the first value is greater than the second
     */
    static anyDeepGreaterThan(a: any, b: any): boolean;
}
//# sourceMappingURL=ArrayValues.d.ts.map