"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BinaryBufferDataResolver = void 0;
const defined_1 = require("../base/defined");
const defaultValue_1 = require("../base/defaultValue");
const ResourceError_1 = require("../base/ResourceError");
/**
 * A class for resolving binary buffer data.
 */
class BinaryBufferDataResolver {
    /**
     * Resolves the buffer data that is defined in the given structure.
     *
     * It receives a `BinaryBufferStructure` that contains the
     * `BufferObject` and `BufferView` definitions, resolves the
     * data from the buffer URIs using the given resource resolver,
     * and returns a `BinaryBufferData` that contains the actual
     * binary buffer data.
     *
     * The given `binaryBuffer` will be used as the buffer data
     * for any buffer that does not have a URI (intended for
     * binary subtree files))
     *
     * @param binaryBufferStructure - The `BinaryBufferStructure`
     * @param binaryBuffer - The optional binary buffer
     * @param resourceResolver - The `ResourceResolver`
     * @returns The `BinaryBufferData`
     * @throws ResourceError If the data could not be resolved
     */
    static async resolve(binaryBufferStructure, binaryBuffer, resourceResolver) {
        // Obtain the buffer data objects: One `Buffer` for
        // each `BufferObject`
        const buffersData = [];
        const buffers = (0, defaultValue_1.defaultValue)(binaryBufferStructure.buffers, []);
        for (const buffer of buffers) {
            if (!(0, defined_1.defined)(buffer.uri)) {
                if (!(0, defined_1.defined)(binaryBuffer)) {
                    throw new ResourceError_1.ResourceError("Expected a binary buffer, but got undefined");
                }
                buffersData.push(binaryBuffer);
            }
            else {
                //console.log("Obtaining buffer data from " + buffer.uri);
                const bufferData = await resourceResolver.resolveData(buffer.uri);
                if (!(0, defined_1.defined)(bufferData)) {
                    const message = `Could not resolve buffer ${buffer.uri}`;
                    throw new ResourceError_1.ResourceError(message);
                }
                buffersData.push(bufferData);
            }
        }
        // Obtain the buffer view data objects: One `Buffer` for
        // each `BufferView`
        const bufferViewsData = [];
        const bufferViews = (0, defaultValue_1.defaultValue)(binaryBufferStructure.bufferViews, []);
        for (const bufferView of bufferViews) {
            const bufferData = buffersData[bufferView.buffer];
            const start = bufferView.byteOffset;
            const end = start + bufferView.byteLength;
            const bufferViewData = bufferData.subarray(start, end);
            bufferViewsData.push(bufferViewData);
        }
        const binarybBufferData = {
            buffersData: buffersData,
            bufferViewsData: bufferViewsData,
        };
        return binarybBufferData;
    }
}
exports.BinaryBufferDataResolver = BinaryBufferDataResolver;
