"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PropertyModels = void 0;
const defined_1 = require("../base/defined");
const defaultValue_1 = require("../base/defaultValue");
const StringPropertyModel_1 = require("./StringPropertyModel");
const BooleanPropertyModel_1 = require("./BooleanPropertyModel");
const NumericPropertyModel_1 = require("./NumericPropertyModel");
const NumericArrayPropertyModel_1 = require("./NumericArrayPropertyModel");
const StringArrayPropertyModel_1 = require("./StringArrayPropertyModel");
const BooleanArrayPropertyModel_1 = require("./BooleanArrayPropertyModel");
const NumericBuffers_1 = require("./NumericBuffers");
/**
 * Methods related to `PropertyModel` instances
 *
 * @internal
 */
class PropertyModels {
    /**
     * Creates a `PropertyModel` for the specified property in
     * the given `BinaryPropertyTable`.
     *
     * This assumes that the input is structurally valid, as
     * determined with the `BinaryPropertyTableValidator`.
     *
     * This will determine the type of the property and access its
     * associated data (i.e. the required buffer views data) from
     * the given `BinaryPropertyTable`. For each type of property,
     * this will return a matching implementation of the
     * `PropertyModel` interface.
     *
     * @param binaryPropertyTable - The `BinaryPropertyTable`
     * @param propertyId - The property ID
     * @returns The `PropertyModel`
     */
    static createPropertyModel(binaryPropertyTable, propertyId) {
        // Obtain the `ClassProperty`
        const metadataClass = binaryPropertyTable.metadataClass;
        const classProperties = metadataClass.properties;
        const classProperty = classProperties[propertyId];
        // Obtain the `PropertyTableProperty`
        const propertyTable = binaryPropertyTable.propertyTable;
        const propertyTableProperties = propertyTable.properties;
        const propertyTableProperty = propertyTableProperties[propertyId];
        // Obtain the required buffers from the binary data:
        const binaryBufferData = binaryPropertyTable.binaryBufferData;
        const bufferViewsData = binaryBufferData.bufferViewsData;
        // Obtain the `values` buffer view data
        const valuesBufferViewIndex = propertyTableProperty.values;
        const valuesBufferViewData = bufferViewsData[valuesBufferViewIndex];
        // Obtain the `arrayOffsets` buffer view data
        const arrayOffsetsBufferViewIndex = propertyTableProperty.arrayOffsets;
        let arrayOffsetsBufferViewData = undefined;
        if ((0, defined_1.defined)(arrayOffsetsBufferViewIndex)) {
            arrayOffsetsBufferViewData =
                bufferViewsData[arrayOffsetsBufferViewIndex];
        }
        const arrayOffsetType = (0, defaultValue_1.defaultValue)(propertyTableProperty.arrayOffsetType, "UINT32");
        // Obtain the `stringOffsets` buffer view data
        const stringOffsetsBufferViewIndex = propertyTableProperty.stringOffsets;
        let stringOffsetsBufferViewData = undefined;
        if ((0, defined_1.defined)(stringOffsetsBufferViewIndex)) {
            stringOffsetsBufferViewData =
                bufferViewsData[stringOffsetsBufferViewIndex];
        }
        const stringOffsetType = (0, defaultValue_1.defaultValue)(propertyTableProperty.stringOffsetType, "UINT32");
        // Determine the `enumValueType` of the property
        const enumType = classProperty.enumType;
        let enumValueType = undefined;
        if ((0, defined_1.defined)(enumType)) {
            const binaryEnumInfo = binaryPropertyTable.binaryEnumInfo;
            const enumValueTypes = binaryEnumInfo.enumValueTypes;
            enumValueType = (0, defaultValue_1.defaultValue)(enumValueTypes[enumType], "UINT16");
        }
        // Create the `PropertyModel` implementation that matches
        // the type of the property
        const type = classProperty.type;
        const componentType = classProperty.componentType;
        const count = classProperty.count;
        const isArray = classProperty.array === true;
        if (isArray) {
            if (type === "STRING") {
                const propertyModel = new StringArrayPropertyModel_1.StringArrayPropertyModel(valuesBufferViewData, arrayOffsetsBufferViewData, arrayOffsetType, stringOffsetsBufferViewData, stringOffsetType, count);
                return propertyModel;
            }
            if (type === "BOOLEAN") {
                const propertyModel = new BooleanArrayPropertyModel_1.BooleanArrayPropertyModel(valuesBufferViewData, arrayOffsetsBufferViewData, arrayOffsetType, count);
                return propertyModel;
            }
            if (type === "ENUM") {
                const propertyModel = new NumericArrayPropertyModel_1.NumericArrayPropertyModel(type, valuesBufferViewData, enumValueType, arrayOffsetsBufferViewData, arrayOffsetType, count);
                return propertyModel;
            }
            // The 'type' must be a numeric (array) type here
            const propertyModel = new NumericArrayPropertyModel_1.NumericArrayPropertyModel(type, valuesBufferViewData, componentType, arrayOffsetsBufferViewData, arrayOffsetType, count);
            return propertyModel;
        }
        // The property must be a non-array property here:
        if (type === "STRING") {
            const propertyModel = new StringPropertyModel_1.StringPropertyModel(valuesBufferViewData, stringOffsetsBufferViewData, stringOffsetType);
            return propertyModel;
        }
        if (type === "BOOLEAN") {
            const propertyModel = new BooleanPropertyModel_1.BooleanPropertyModel(valuesBufferViewData);
            return propertyModel;
        }
        if (type === "ENUM") {
            const propertyModel = new NumericPropertyModel_1.NumericPropertyModel(type, valuesBufferViewData, enumValueType);
            return propertyModel;
        }
        // The property must be a (non-array) numeric property here
        const propertyModel = new NumericPropertyModel_1.NumericPropertyModel(type, valuesBufferViewData, componentType);
        return propertyModel;
    }
    /**
     * Returns the 'slice' information that is given by an offsets
     * buffer or a fixed number.
     *
     * This returns `{ offset, length }` for the `arrayOffsets` or
     * `stringOffsets` of a property, for a given index.
     *
     * When the given `count` is defined, then the result will
     * just be `{ index * count, count }`.
     *
     * Otherwise, the result will be `{ offset, length }`, where `offset`
     * is the offset that is read from the given buffer at index `index`,
     * and `length` is `offset[index+1] - offset[index]`.
     *
     * @param index - The index
     * @param offsetsBuffer - The offsets
     * @param offsetType - The `componentType` for the offsets
     * @param count - The count
     * @returns The slice information
     */
    static computeSlice(index, offsetsBuffer, offsetType, count) {
        if ((0, defined_1.defined)(count)) {
            return {
                offset: index * count,
                length: count,
            };
        }
        const offset = NumericBuffers_1.NumericBuffers.getNumericFromBuffer(offsetsBuffer, index, offsetType);
        const nextOffset = NumericBuffers_1.NumericBuffers.getNumericFromBuffer(offsetsBuffer, index + 1, offsetType);
        const length = nextOffset - offset;
        return {
            offset: offset,
            length: length,
        };
    }
}
exports.PropertyModels = PropertyModels;
