"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PropertyTableModel = void 0;
const defaultValue_1 = require("../base/defaultValue");
const BinaryMetadataEntityModel_1 = require("./BinaryMetadataEntityModel");
const PropertyModels_1 = require("./PropertyModels");
const MetadataEntityModels_1 = require("../metadata/MetadataEntityModels");
const MetadataError_1 = require("../metadata/MetadataError");
/**
 * Implementation of a model for a property table that is backed
 * by binary data.
 *
 * @internal
 */
class PropertyTableModel {
    constructor(binaryPropertyTable) {
        /**
         * A mapping from property IDs to the `PropertyModel`
         * instances that provide the property values. These
         * are the "columns" of the table
         */
        this._propertyIdToModel = {};
        this._binaryPropertyTable = binaryPropertyTable;
        // Initialize the `PropertyModel` instances for
        // the property table properties
        const propertyTable = this._binaryPropertyTable.propertyTable;
        const propertyTableProperties = (0, defaultValue_1.defaultValue)(propertyTable.properties, {});
        for (const propertyId of Object.keys(propertyTableProperties)) {
            const propertyModel = PropertyModels_1.PropertyModels.createPropertyModel(this._binaryPropertyTable, propertyId);
            this._propertyIdToModel[propertyId] = propertyModel;
        }
        const metadataClass = binaryPropertyTable.metadataClass;
        this._semanticToPropertyId =
            MetadataEntityModels_1.MetadataEntityModels.computeSemanticToPropertyIdMapping(metadataClass);
    }
    /**
     * Returns the `MetadataEntityModel` that corresponds to the
     * row of the table with the given index.
     *
     * @param index - The index (i.e. the table row)
     * @returns The `MetdataEntityModel`
     * @throws MetadataError If the index is out of range
     */
    getMetadataEntityModel(index) {
        const propertyTable = this._binaryPropertyTable.propertyTable;
        const count = propertyTable.count;
        if (index < 0 || index >= count) {
            const message = `The index must be in [0,${count}), but is ${index}`;
            throw new MetadataError_1.MetadataError(message);
        }
        const semanticToPropertyId = this._semanticToPropertyId;
        const metadataEntityModel = new BinaryMetadataEntityModel_1.BinaryMetadataEntityModel(this, index, semanticToPropertyId);
        return metadataEntityModel;
    }
    /**
     * Returns the `ClassProperty` that defines the structure of the
     * property with the given ID, or `undefined` if this table was
     * created for a `MetadataClass` that does not define this property.
     *
     * @param propertyId - The property ID
     * @returns The `ClassProperty`
     */
    getClassProperty(propertyId) {
        const binaryPropertyTable = this._binaryPropertyTable;
        const metadataClass = binaryPropertyTable.metadataClass;
        const classProperties = (0, defaultValue_1.defaultValue)(metadataClass.properties, {});
        return classProperties[propertyId];
    }
    /**
     * Returns the `PropertyTableProperty` that defines the structure of the
     * property with the given ID, or `undefined` if this table was
     * created for a `PropertyTable` that does not define this property.
     *
     * @param propertyId - The property ID
     * @returns The `PropertyTableProperty`
     */
    getPropertyTableProperty(propertyId) {
        const binaryPropertyTable = this._binaryPropertyTable;
        const propertyTable = binaryPropertyTable.propertyTable;
        const propertyTableProperties = (0, defaultValue_1.defaultValue)(propertyTable.properties, {});
        return propertyTableProperties[propertyId];
    }
    /**
     * Returns the `PropertyModel` for the property with the given ID.
     * This is the "column" of the table that contains the property
     * data. Returns `undefined` if this table was created for
     * a `MetadataClass` that does not define this property.
     *
     * @param propertyId - The property ID
     * @returns The `PropertyModel`
     */
    getPropertyModel(propertyId) {
        return this._propertyIdToModel[propertyId];
    }
}
exports.PropertyTableModel = PropertyTableModel;
