/**
 * Morton Order (aka Z-Order Curve) helper functions.
 * @see {@link https://en.wikipedia.org/wiki/Z-order_curve}
 *
 * @internal
 */
export declare class MortonOrder {
    /**
     * Inserts one 0 bit of spacing between a number's bits. This is the opposite of removeOneSpacing.
     *
     * Example:
     *  input: 6
     *  input (binary):  110
     *  output (binary): 10100
     *                    ^ ^ (added)
     *  output: 20
     *
     * @param v - A 16-bit unsigned integer.
     * @returns A 32-bit unsigned integer.
     * @see {@link https://fgiesen.wordpress.com/2009/12/13/decoding-morton-codes/}
     */
    private static insertOneSpacing;
    /**
     * Inserts two 0 bits of spacing between a number's bits. This is the opposite of removeTwoSpacing.
     *
     * Example:
     *  input: 6
     *  input (binary):  110
     *  output (binary): 1001000
     *                    ^^ ^^ (added)
     *  output: 72
     *
     * @internal
     * @param v - A 10-bit unsigned integer.
     * @returns A 30-bit unsigned integer.
     * @see {@link https://fgiesen.wordpress.com/2009/12/13/decoding-morton-codes/}
     */
    private static insertTwoSpacing;
    /**
     * Removes one bit of spacing between bits. This is the opposite of insertOneSpacing.
     *
     * Example:
     *  input: 20
     *  input (binary):  10100
     *                    ^ ^ (removed)
     *  output (binary): 110
     *  output: 6
     *
     * @param v - A 32-bit unsigned integer.
     * @returns A 16-bit unsigned integer.
     * @see {@link https://fgiesen.wordpress.com/2009/12/13/decoding-morton-codes/}
     */
    private static removeOneSpacing;
    /**
     * Removes two bits of spacing between bits. This is the opposite of insertTwoSpacing.
     *
     * Example:
     *  input: 72
     *  input (binary):  1001000
     *                    ^^ ^^ (removed)
     *  output (binary): 110
     *  output: 6
     *
     * @param v - A 30-bit unsigned integer.
     * @returns A 10-bit unsigned integer.
     * @see {@link https://fgiesen.wordpress.com/2009/12/13/decoding-morton-codes/}
     */
    private static removeTwoSpacing;
    /**
     * Computes the Morton index from 2D coordinates. This is equivalent to interleaving their bits.
     * The inputs must be 16-bit unsigned integers (resulting in 32-bit Morton index) due to 32-bit bitwise operator limitation in JavaScript.
     *
     * @param x - The X coordinate in the range [0, (2^16)-1].
     * @param y - The Y coordinate in the range [0, (2^16)-1].
     * @returns The Morton index.
     * @internal
     */
    static encode2D(x: number, y: number): number;
    /**
     * Computes the 2D coordinates from a Morton index. This is equivalent to deinterleaving their bits.
     * The input must be a 32-bit unsigned integer (resulting in 16 bits per coordinate) due to 32-bit bitwise operator limitation in JavaScript.
     *
     * @param mortonIndex - The Morton index in the range [0, (2^32)-1].
     * @param result - The array onto which to store the result.
     * @returns An array containing the 2D coordinates correspoding to the Morton index.
     * @internal
     */
    static decode2D(mortonIndex: number, result: number[]): number[];
    /**
     * Computes the Morton index from 3D coordinates. This is equivalent to interleaving their bits.
     * The inputs must be 10-bit unsigned integers (resulting in 30-bit Morton index) due to 32-bit bitwise operator limitation in JavaScript.
     *
     * @param x - The X coordinate in the range [0, (2^10)-1].
     * @param y - The Y coordinate in the range [0, (2^10)-1].
     * @param z - The Z coordinate in the range [0, (2^10)-1].
     * @returns The Morton index.
     * @internal
     */
    static encode3D(x: number, y: number, z: number): number;
    /**
     * Computes the 3D coordinates from a Morton index. This is equivalent to deinterleaving their bits.
     * The input must be a 30-bit unsigned integer (resulting in 10 bits per coordinate) due to 32-bit bitwise operator limitation in JavaScript.
     *
     * @param mortonIndex - The Morton index in the range [0, (2^30)-1].
     * @param result - The array onto which to store the result.
     * @returns An array containing the 3D coordinates corresponding to the Morton index.
     * @internal
     */
    static decode3D(mortonIndex: number, result: number[]): number[];
}
//# sourceMappingURL=MortonOrder.d.ts.map