"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SubtreeInfos = void 0;
const defined_1 = require("../base/defined");
const bufferToJson_1 = require("../base/bufferToJson");
const ResourceError_1 = require("../base/ResourceError");
const SubtreeInfo_1 = require("./SubtreeInfo");
const AvailabilityInfos_1 = require("./AvailabilityInfos");
const ImplicitTilingError_1 = require("./ImplicitTilingError");
const BinaryBufferDataResolver_1 = require("../binary/BinaryBufferDataResolver");
/**
 * Methods to create `SubtreeInfo` instances.
 */
class SubtreeInfos {
    /**
     * Creates a new `SubtreeInfo` from the given binary subtree data.
     *
     * This method assumes that the given binary data is consistent
     * and valid. This can be checked with the `SubtreeValidator`
     * class.
     *
     * @param input - The whole buffer of a binary subtree file
     * @param implicitTiling - The `TileImplicitTiling` that
     * defines the expected structure of the subtree data
     * @param resourceResolver - The `ResourceResolver` that
     * will be used to resolve buffer URIs
     * @returns A promise with the `SubtreeInfo`
     * @throws An ImplicitTilingError when the subtree JSON could
     * not be parsed, or there was a buffer without a URI
     * and no binary buffer was given, or one of the requested
     * buffers could not be resolved.
     */
    static async createFromBuffer(input, implicitTiling, resourceResolver) {
        const headerByteLength = 24;
        const jsonByteLength = input.readBigUint64LE(8);
        const binaryByteLength = input.readBigUint64LE(16);
        // Extract the JSON data
        const jsonStartByteOffset = headerByteLength;
        const jsonEndByteOffset = jsonStartByteOffset + Number(jsonByteLength);
        const jsonBuffer = input.subarray(jsonStartByteOffset, jsonEndByteOffset);
        let subtreeJson;
        let subtree;
        try {
            subtreeJson = (0, bufferToJson_1.bufferToJson)(jsonBuffer);
            subtree = subtreeJson;
        }
        catch (error) {
            throw new ImplicitTilingError_1.ImplicitTilingError("Could not parse subtree JSON data");
        }
        // Extract the binary buffer
        const binaryStartByteOffset = jsonEndByteOffset;
        const binaryEndByteOffset = binaryStartByteOffset + Number(binaryByteLength);
        const binaryBufferSlice = input.subarray(binaryStartByteOffset, binaryEndByteOffset);
        const binaryBuffer = binaryBufferSlice.length > 0 ? binaryBufferSlice : undefined;
        return SubtreeInfos.create(subtree, binaryBuffer, implicitTiling, resourceResolver);
    }
    /**
     * Creates a new `SubtreeInfo` from the given `Subtree` object
     * and the (optional) binary buffer.
     *
     * This method assumes that the given data is consistent
     * and valid. This can be checked with the `SubtreeValidator`
     * class.
     *
     * @param subtree - The `Subtree` object
     * @param binaryBuffer - The optional binary buffer
     * @param implicitTiling - The `TileImplicitTiling` that
     * defines the expected structure of the subtree data
     * @param resourceResolver - The `ResourceResolver` that
     * will be used to resolve buffer URIs
     * @returns A promise with the `SubtreeInfo`
     * @throws A ImplicitTilingError when there was a buffer without
     * a URI and no binary buffer was given, or the requested buffer
     * data could not be resolved.
     */
    static async create(subtree, binaryBuffer, implicitTiling, resourceResolver) {
        const binaryBufferStructure = {
            buffers: subtree.buffers,
            bufferViews: subtree.bufferViews,
        };
        let binaryBufferData;
        try {
            binaryBufferData = await BinaryBufferDataResolver_1.BinaryBufferDataResolver.resolve(binaryBufferStructure, binaryBuffer, resourceResolver);
        }
        catch (error) {
            if (error instanceof ResourceError_1.ResourceError) {
                const message = `Could not read subtree data: ${error.message}`;
                throw new ImplicitTilingError_1.ImplicitTilingError(message);
            }
            throw error;
        }
        const bufferViewsData = binaryBufferData.bufferViewsData;
        // Create the `AvailabilityInfo` for the tile availability
        const tileAvailability = subtree.tileAvailability;
        const tileAvailabilityInfo = AvailabilityInfos_1.AvailabilityInfos.createTileOrContent(tileAvailability, bufferViewsData, implicitTiling);
        // Create the `AvailabilityInfo` objects, one for
        // each content availability
        const contentAvailabilityInfos = [];
        const contentAvailabilities = subtree.contentAvailability;
        if ((0, defined_1.defined)(contentAvailabilities)) {
            for (const contentAvailability of contentAvailabilities) {
                const contentAvailabilityInfo = AvailabilityInfos_1.AvailabilityInfos.createTileOrContent(contentAvailability, bufferViewsData, implicitTiling);
                contentAvailabilityInfos.push(contentAvailabilityInfo);
            }
        }
        // Create the `AvailabilityInfo` for the child subtree availability
        const childSubtreeAvailability = subtree.childSubtreeAvailability;
        const childSubtreeAvailabilityInfo = AvailabilityInfos_1.AvailabilityInfos.createChildSubtree(childSubtreeAvailability, bufferViewsData, implicitTiling);
        const result = new SubtreeInfo_1.SubtreeInfo(tileAvailabilityInfo, contentAvailabilityInfos, childSubtreeAvailabilityInfo);
        return result;
    }
}
exports.SubtreeInfos = SubtreeInfos;
