"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.UnzippingResourceResolver = void 0;
const zlib_1 = __importDefault(require("zlib"));
const ResourceTypes_1 = require("./ResourceTypes");
/**
 * Implementation of a `ResourceResolver` that obtains the resource
 * data from a delegate, and unzips the data if necessary.
 *
 * @internal (Instantiated by the `ResourceResolvers` class)
 */
class UnzippingResourceResolver {
    constructor(delegate) {
        this._delegate = delegate;
    }
    resolveUri(uri) {
        return this._delegate.resolveUri(uri);
    }
    async resolveData(uri) {
        const delegateData = await this._delegate.resolveData(uri);
        if (delegateData === null) {
            return null;
        }
        const isGzipped = ResourceTypes_1.ResourceTypes.isGzipped(delegateData);
        if (!isGzipped) {
            return delegateData;
        }
        const data = zlib_1.default.gunzipSync(delegateData);
        return data;
    }
    async resolveDataPartial(uri, maxBytes) {
        const partialDelegateData = await this._delegate.resolveDataPartial(uri, maxBytes);
        if (partialDelegateData === null) {
            return null;
        }
        const isGzipped = ResourceTypes_1.ResourceTypes.isGzipped(partialDelegateData);
        if (!isGzipped) {
            return partialDelegateData;
        }
        const fullDelegateData = await this._delegate.resolveData(uri);
        if (fullDelegateData === null) {
            return null;
        }
        const data = zlib_1.default.gunzipSync(fullDelegateData);
        return data;
    }
    derive(uri) {
        return new UnzippingResourceResolver(this._delegate.derive(uri));
    }
}
exports.UnzippingResourceResolver = UnzippingResourceResolver;
