"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GltfValidator = void 0;
const defined_1 = require("../base/defined");
const ContentValidationIssues_1 = require("../issues/ContentValidationIssues");
// eslint-disable-next-line @typescript-eslint/no-var-requires
const validator = require("gltf-validator");
/**
 * A thin wrapper around the `gltf-validator`, implementing the
 * `Validator` interface for glTF- and GLB data that is given
 * in a Buffer.
 */
class GltfValidator {
    /**
     * Creates a `ValidationIssue` object for the given 'message' object
     * that appears in the output of the glTF validator.
     *
     * @param gltfMessage - The glTF validator message
     * @returns The `ValidationIssue`
     */
    static createValidationIssueFromGltfMessage(gltfMessage) {
        const path = gltfMessage.pointer;
        const message = gltfMessage.message;
        if (gltfMessage.severity == 0) {
            const issue = ContentValidationIssues_1.ContentValidationIssues.CONTENT_VALIDATION_ERROR(path, message);
            return issue;
        }
        const issue = ContentValidationIssues_1.ContentValidationIssues.CONTENT_VALIDATION_WARNING(path, message);
        return issue;
    }
    async validateObject(uri, input, context) {
        const resourceResolver = context.getResourceResolver();
        let gltfResult = undefined;
        try {
            gltfResult = await validator.validateBytes(input, {
                uri: uri,
                externalResourceFunction: (gltfUri) => {
                    const resolvedDataPromise = resourceResolver.resolveData(gltfUri);
                    return resolvedDataPromise.then((resolvedData) => {
                        if (!(0, defined_1.defined)(resolvedData)) {
                            throw "Could not resolve data from " + gltfUri;
                        }
                        return resolvedData;
                    });
                },
            });
        }
        catch (error) {
            const message = `Content ${uri} caused internal validation error: ${error}`;
            const issue = ContentValidationIssues_1.ContentValidationIssues.CONTENT_VALIDATION_ERROR(uri, message);
            context.addIssue(issue);
            return false;
        }
        // If there are any errors, then summarize ALL issues from the glTF
        // validation as 'internal issues' in a CONTENT_VALIDATION_ERROR
        if (gltfResult.issues.numErrors > 0) {
            const path = uri;
            const message = `Content ${uri} caused validation errors`;
            const issue = ContentValidationIssues_1.ContentValidationIssues.CONTENT_VALIDATION_ERROR(path, message);
            for (const gltfMessage of gltfResult.issues.messages) {
                //console.log(gltfMessage);
                const cause = GltfValidator.createValidationIssueFromGltfMessage(gltfMessage);
                issue.addCause(cause);
            }
            context.addIssue(issue);
            // Consider the object to be invalid
            return false;
        }
        // If there are any warnings, then summarize them in a
        // CONTENT_VALIDATION_WARNING, but still consider the
        // object to be valid.
        if (gltfResult.issues.numWarnings > 0) {
            const path = uri;
            const message = `Content ${uri} caused validation warnings`;
            const issue = ContentValidationIssues_1.ContentValidationIssues.CONTENT_VALIDATION_WARNING(path, message);
            for (const gltfMessage of gltfResult.issues.messages) {
                //console.log(gltfMessage);
                const cause = GltfValidator.createValidationIssueFromGltfMessage(gltfMessage);
                issue.addCause(cause);
            }
            context.addIssue(issue);
        }
        return true;
    }
}
exports.GltfValidator = GltfValidator;
