"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ExplicitTraversedTile = void 0;
const defined_1 = require("../base/defined");
const ImplicitTileTraversal_1 = require("./ImplicitTileTraversal");
const MetadataEntityModels_1 = require("../metadata/MetadataEntityModels");
const ImplicitTilingError_1 = require("../implicitTiling/ImplicitTilingError");
const ImplicitTilings_1 = require("../implicitTiling/ImplicitTilings");
/**
 * An implementation of a `TraversedTile` that reflects a tile
 * that actually appears as a JSON representation in the tileset.
 */
class ExplicitTraversedTile {
    constructor(tile, path, level, parent, schema, resourceResolver) {
        // TODO For debugging
        this.toString = () => {
            return `ExplicitTraversedTile, level ${this.level}, path ${this.path}`;
        };
        this._tile = tile;
        this._path = path;
        this._level = level;
        this._parent = parent;
        this._schema = schema;
        this._resourceResolver = resourceResolver;
    }
    asTile() {
        const tile = this._tile;
        const schema = this._schema;
        const metadata = tile.metadata;
        const boundingVolume = tile.boundingVolume;
        let transform = tile.transform;
        let refine = tile.refine;
        let geometricError = tile.geometricError;
        if ((0, defined_1.defined)(metadata) && (0, defined_1.defined)(schema)) {
            let metadataEntityModel = undefined;
            try {
                metadataEntityModel = MetadataEntityModels_1.MetadataEntityModels.create(schema, metadata);
            }
            catch (error) {
                // Errors from creating the entity model should have been
                // prevented by the metadata- and schema validation:
                const message = `Error while traversing tileset: ${error}`;
                throw new ImplicitTilingError_1.ImplicitTilingError(message);
            }
            if ((0, defined_1.defined)(metadataEntityModel)) {
                // Apply the semantic-based overrides from the metadata
                const semanticBoundingBox = metadataEntityModel.getPropertyValueBySemantic("TILE_BOUNDING_BOX");
                if ((0, defined_1.defined)(semanticBoundingBox)) {
                    boundingVolume.box = semanticBoundingBox;
                }
                const semanticBoundingRegion = metadataEntityModel.getPropertyValueBySemantic("TILE_BOUNDING_REGION");
                if ((0, defined_1.defined)(semanticBoundingRegion)) {
                    boundingVolume.region = semanticBoundingRegion;
                }
                const semanticBoundingSphere = metadataEntityModel.getPropertyValueBySemantic("TILE_BOUNDING_SPHERE");
                if ((0, defined_1.defined)(semanticBoundingSphere)) {
                    boundingVolume.sphere = semanticBoundingSphere;
                }
                const semanticGeometricError = metadataEntityModel.getPropertyValueBySemantic("TILE_GEOMETRIC_ERROR");
                if ((0, defined_1.defined)(semanticGeometricError)) {
                    geometricError = semanticGeometricError;
                }
                const semanticRefine = metadataEntityModel.getPropertyValueBySemantic("TILE_REFINE");
                if (semanticRefine === 0) {
                    refine = "ADD";
                }
                else if (semanticRefine === 1) {
                    refine = "REPLACE";
                }
                const semanticTransform = metadataEntityModel.getPropertyValueBySemantic("TILE_TRANSFORM");
                if ((0, defined_1.defined)(semanticTransform)) {
                    transform = semanticTransform;
                }
            }
        }
        const viewerRequestVolume = tile.viewerRequestVolume;
        const content = tile.content;
        const contents = tile.contents;
        const children = tile.children;
        const implicitTiling = tile.implicitTiling;
        const extensions = tile.extensions;
        const extras = tile.extras;
        return {
            boundingVolume: boundingVolume,
            viewerRequestVolume: viewerRequestVolume,
            geometricError: geometricError,
            refine: refine,
            transform: transform,
            content: content,
            contents: contents,
            children: children,
            metadata: metadata,
            implicitTiling: implicitTiling,
            extensions: extensions,
            extras: extras,
        };
    }
    get path() {
        return this._path;
    }
    get level() {
        return this._level;
    }
    getParent() {
        return this._parent;
    }
    async getChildren() {
        const implicitTiling = this._tile.implicitTiling;
        if ((0, defined_1.defined)(implicitTiling)) {
            const children = await ImplicitTileTraversal_1.ImplicitTileTraversal.createTraversedChildren(implicitTiling, this, this._resourceResolver);
            return children;
        }
        if (!(0, defined_1.defined)(this._tile.children)) {
            return [];
        }
        const children = this._tile.children;
        const childLevel = this._level + 1;
        const traversedChildren = [];
        for (let i = 0; i < children.length; i++) {
            const child = children[i];
            const childPath = this.path + "/children/" + i;
            const traversedChild = new ExplicitTraversedTile(child, childPath, childLevel, this, this._schema, this._resourceResolver);
            traversedChildren.push(traversedChild);
        }
        return traversedChildren;
    }
    getContents() {
        if ((0, defined_1.defined)(this._tile.content)) {
            return [this._tile.content];
        }
        if ((0, defined_1.defined)(this._tile.contents)) {
            return this._tile.contents;
        }
        return [];
    }
    getSubtreeUri() {
        const implicitTiling = this._tile.implicitTiling;
        if ((0, defined_1.defined)(implicitTiling)) {
            const rootCoordinates = ImplicitTilings_1.ImplicitTilings.createRootCoordinates(implicitTiling);
            const subtreeUri = ImplicitTilings_1.ImplicitTilings.substituteTemplateUri(implicitTiling.subdivisionScheme, implicitTiling.subtrees.uri, rootCoordinates);
            return subtreeUri;
        }
        return undefined;
    }
    getImplicitTiling() {
        return this._tile.implicitTiling;
    }
    getMetadata() {
        return this._tile.metadata;
    }
}
exports.ExplicitTraversedTile = ExplicitTraversedTile;
