import { ResourceResolver } from "../io/ResourceResolver";
import { TraversedTile } from "./TraversedTile";
import { ExplicitTraversedTile } from "./ExplicitTraversedTile";
import { TileImplicitTiling } from "../structure/TileImplicitTiling";
import { TreeCoordinates } from "../implicitTiling/TreeCoordinates";
import { SubtreeInfo } from "../implicitTiling/SubtreeInfo";
/**
 * Methods related to the traversal of implicit tilesets.
 *
 * @internal
 */
export declare class ImplicitTileTraversal {
    /**
     * Create the traversed children for the given explicit traversed tile.
     *
     * This method will be called from `ExplicitTraversedTile` instances
     * when the contain `implicitTiling` information, in order to create
     * the traversed children.
     *
     * The children will then be a single-element array that contains the
     * root node of the implicit tileset, as an `ImplicitTraversedTile`.
     *
     * @param implicitTiling - The `TileImplicitTiling`
     * @param parent - The `ExplicitTraversedTile`
     * @param resourceResolver - The `ResourceResolver` that
     * will be used e.g. for subtree files
     * @returns The traversed children
     * @throws ImplicitTilingError If the input was structurally invalid
     */
    static createTraversedChildren(implicitTiling: TileImplicitTiling, parent: ExplicitTraversedTile, resourceResolver: ResourceResolver): Promise<TraversedTile[]>;
    /**
     * Creates the root node for the traversal of an implicit quadtree.
     *
     * @param implicitTiling - The `TileImplicitTiling`
     * @param parent - The `ExplicitTraversedTile`
     * @param resourceResolver - The `ResourceResolver` that
     * will be used e.g. for subtree files
     * @returns The root of an implicit quadtree
     * @throws ImplicitTilingError If the input was structurally invalid
     */
    private static createImplicitQuadtreeRoot;
    /**
     * Creates the root node for the traversal of an implicit octree.
     *
     * @param implicitTiling - The `TileImplicitTiling`
     * @param parent - The `ExplicitTraversedTile`
     * @param resourceResolver - The `ResourceResolver` that
     * will be used e.g. for subtree files
     * @returns The root of an implicit octree
     * @throws ImplicitTilingError If the input was structurally invalid
     */
    private static createImplicitOctreeRoot;
    /**
     * Resolve the `SubtreeInfo` for the subtree with the given root coordinates.
     *
     * This will substitute the given coordinates into the subtree template
     * URI from the given implicit tiling object. Then it will attempt to load
     * the subtree data from this URI. The resulting data will be used to
     * construct the `SubtreeInfo` object.
     *
     * @param implicitTiling - The `TileImplicitTiling`
     * @param resourceResolver - The `ResourceResolver` for the subtree
     * files and buffers
     * @param coordinates - The root coordinates of the subtree
     * @returns The `SubtreeInfo`
     * @throws ImplicitTilingError If the input was structurally invalid
     */
    static resolveSubtreeInfo(implicitTiling: TileImplicitTiling, resourceResolver: ResourceResolver, coordinates: TreeCoordinates): Promise<SubtreeInfo>;
}
//# sourceMappingURL=ImplicitTileTraversal.d.ts.map