"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ImplicitTraversedTile = void 0;
const defined_1 = require("../base/defined");
const ImplicitTileTraversal_1 = require("./ImplicitTileTraversal");
const BoundingVolumeDerivation_1 = require("./cesium/BoundingVolumeDerivation");
const ImplicitTilingError_1 = require("../implicitTiling/ImplicitTilingError");
const ImplicitTilings_1 = require("../implicitTiling/ImplicitTilings");
/**
 * An implementation of a `TraversedTile` that represents a tile
 * within an implicit tileset during its traversal.
 */
class ImplicitTraversedTile {
    constructor(implicitTiling, resourceResolver, root, path, subtreeInfo, globalLevel, globalCoordinate, rootCoordinate, localCoordinate, parent) {
        // TODO For debugging
        this.toString = () => {
            return (`ImplicitTraversedTile, ` +
                `level ${this._globalLevel}, ` +
                `global: ${this._globalCoordinate}, ` +
                `root: ${this._rootCoordinate}, ` +
                `local: ${this._localCoordinate}`
            //`path ${this.path}`
            );
        };
        this._implicitTiling = implicitTiling;
        this._resourceResolver = resourceResolver;
        this._root = root;
        this._path = path;
        this._subtreeInfo = subtreeInfo;
        this._globalLevel = globalLevel;
        this._globalCoordinate = globalCoordinate;
        this._rootCoordinate = rootCoordinate;
        this._localCoordinate = localCoordinate;
        this._parent = parent;
    }
    asTile() {
        // TODO The bounding volume and geometric error
        // may be overridden via semantics!
        const rootTile = this._root.asTile();
        const boundingVolume = BoundingVolumeDerivation_1.BoundingVolumeDerivation.deriveBoundingVolume(rootTile.boundingVolume, this._globalCoordinate.toArray());
        if (!(0, defined_1.defined)(boundingVolume)) {
            // The bounding volume neither contained a region nor a box.
            // This should have been detected by previous validation.
            throw new ImplicitTilingError_1.ImplicitTilingError("Could not subdivide bounding volume");
        }
        const level = this._globalCoordinate.level;
        const geometricError = rootTile.geometricError / Math.pow(2, level);
        const viewerRequestVolume = rootTile.viewerRequestVolume;
        const refine = rootTile.refine;
        const transform = undefined;
        const metadata = undefined;
        const contents = this.getContents();
        const implicitTiling = undefined;
        const extensions = undefined;
        const extras = undefined;
        return {
            boundingVolume: boundingVolume,
            viewerRequestVolume: viewerRequestVolume,
            geometricError: geometricError,
            refine: refine,
            transform: transform,
            metadata: metadata,
            contents: contents,
            implicitTiling: implicitTiling,
            extensions: extensions,
            extras: extras,
        };
    }
    get path() {
        return this._path;
    }
    get level() {
        return this._globalLevel;
    }
    getLocalCoordinate() {
        return this._localCoordinate;
    }
    getGlobalCoordinate() {
        return this._globalCoordinate;
    }
    getParent() {
        return this._parent;
    }
    async getChildren() {
        const localLevel = this._localCoordinate.level;
        if (localLevel === this._implicitTiling.subtreeLevels - 1) {
            const children = await this.createNextSubtreeLevelChildren();
            return children;
        }
        const children = await this.createDirectChildren();
        return children;
    }
    /**
     * Creates the children for this tile at which a new subtree starts.
     *
     * This assumes that this tile is in the last level of the subtree
     * that it belongs to. This method will create one child tile for
     * each available child subtree. These children will be the "local
     * roots" of their respective subtree.
     *
     * @returns The children
     * @throws ImplicitTilingError If the input data was invalid
     */
    async createNextSubtreeLevelChildren() {
        const traversedChildren = [];
        const localChildCoordinates = this._localCoordinate.children();
        for (const localChildCoordinate of localChildCoordinates) {
            const globalChildCoordinate = ImplicitTilings_1.ImplicitTilings.globalizeCoordinates(this._implicitTiling, this._rootCoordinate, localChildCoordinate);
            const childSubtreeAvailability = this._subtreeInfo.getChildSubtreeAvailabilityInfo();
            const childSubtreeAvailable = childSubtreeAvailability.isAvailable(localChildCoordinate.toIndexInLevel());
            if (childSubtreeAvailable) {
                const childSubtreeInfo = await ImplicitTileTraversal_1.ImplicitTileTraversal.resolveSubtreeInfo(this._implicitTiling, this._resourceResolver, globalChildCoordinate);
                const childLocalCoordinate = ImplicitTilings_1.ImplicitTilings.createRootCoordinates(this._implicitTiling);
                // The path is composed from the path of the root and the string
                // representation of the global coordinates of the child
                const coordinateString = ImplicitTilings_1.ImplicitTilings.createString(globalChildCoordinate);
                const childPath = `${this._root.path}/${coordinateString}`;
                const traversedChild = new ImplicitTraversedTile(this._implicitTiling, this._resourceResolver, this._root, childPath, childSubtreeInfo, this._globalLevel + 1, globalChildCoordinate, globalChildCoordinate, childLocalCoordinate, this);
                traversedChildren.push(traversedChild);
            }
        }
        return traversedChildren;
    }
    /**
     * Creates the children for this tile that are still within the same subtree.
     *
     * This assumes that this tile is **NOT** in the last level of the subtree.
     * It will return all children that are marked as available, via the
     * tile availability information in the subtree that this tile belongs to.
     *
     * @returns The children
     * @throws ImplicitTilingError If the input data was invalid
     */
    async createDirectChildren() {
        const tileAvailabilityInfo = this._subtreeInfo.getTileAvailabilityInfo();
        const localChildCoordinates = this._localCoordinate.children();
        const traversedChildren = [];
        for (const localChildCoordinate of localChildCoordinates) {
            const available = tileAvailabilityInfo.isAvailable(localChildCoordinate.toIndex());
            if (available) {
                const globalChildCoordinate = ImplicitTilings_1.ImplicitTilings.globalizeCoordinates(this._implicitTiling, this._rootCoordinate, localChildCoordinate);
                // The path is composed from the path of the root and the string
                // representation of the global coordinates of the child
                const coordinateString = ImplicitTilings_1.ImplicitTilings.createString(globalChildCoordinate);
                const childPath = `${this._root.path}/${coordinateString}`;
                const traversedChild = new ImplicitTraversedTile(this._implicitTiling, this._resourceResolver, this._root, childPath, this._subtreeInfo, this._globalLevel + 1, globalChildCoordinate, this._rootCoordinate, localChildCoordinate, this);
                traversedChildren.push(traversedChild);
            }
        }
        return traversedChildren;
    }
    getContents() {
        const contents = [];
        const contentAvailabilityInfos = this._subtreeInfo.getContentAvailabilityInfos();
        for (const contentAvailabilityInfo of contentAvailabilityInfos) {
            const available = contentAvailabilityInfo.isAvailable(this._localCoordinate.toIndex());
            if (available) {
                // TODO The existence of the root content URI should
                // have been validated. So this could also throw
                // an error if the template URI is not found.
                const templateUri = this._root.asTile().content?.uri;
                if ((0, defined_1.defined)(templateUri)) {
                    const contentUri = ImplicitTilings_1.ImplicitTilings.substituteTemplateUri(this._implicitTiling.subdivisionScheme, templateUri, this._globalCoordinate);
                    // TODO Check semantics!
                    const content = {
                        boundingVolume: undefined,
                        uri: contentUri,
                        metadata: undefined,
                        group: undefined,
                    };
                    contents.push(content);
                }
            }
        }
        return contents;
    }
    getSubtreeUri() {
        const localCoordinate = this._localCoordinate;
        if (localCoordinate.level === 0) {
            const globalCoordinate = this._globalCoordinate;
            const implicitTiling = this._implicitTiling;
            const subtreeUri = ImplicitTilings_1.ImplicitTilings.substituteTemplateUri(implicitTiling.subdivisionScheme, implicitTiling.subtrees.uri, globalCoordinate);
            return subtreeUri;
        }
        return undefined;
    }
    getImplicitTiling() {
        const localCoordinate = this._localCoordinate;
        if (localCoordinate.level === 0) {
            return this._implicitTiling;
        }
    }
    getMetadata() {
        return undefined;
    }
}
exports.ImplicitTraversedTile = ImplicitTraversedTile;
