"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TilesetTraverser = void 0;
const defined_1 = require("../base/defined");
const ExplicitTraversedTile_1 = require("./ExplicitTraversedTile");
/**
 * A class that can traverse the tiles of a tileset.
 */
class TilesetTraverser {
    /**
     * Traverses the tiles in the given tileset.
     *
     * This will traverse the tiles of the given tileset, starting
     * at the root. It will pass all tiles to the given callback,
     * as `TraversedTile` instances.
     *
     * @param tileset - The `Tileset`
     * @param schema - The schema from the `tileset.schema` or the
     * `tileset.schemaUri`. If this is defined, then it is assumed
     * to be a valid schema definition.
     * @param resourceResolver - The `ResourceResolver` that is used to
     * resolve resources for implicit tilesets (subtree files)
     * @param traversalCallback - The `TraversalCallback`
     * @param depthFirst - Whether the traversal should be depth-first
     * @returns A Promise that resolves when the traversal finished
     */
    static async traverse(tileset, schema, resourceResolver, traversalCallback, 
    // eslint-disable-next-line @typescript-eslint/no-inferrable-types
    depthFirst = false) {
        const root = tileset.root;
        if (!(0, defined_1.defined)(root)) {
            return;
        }
        const stack = [];
        const traversedRoot = new ExplicitTraversedTile_1.ExplicitTraversedTile(root, "/root", 0, undefined, schema, resourceResolver);
        stack.push(traversedRoot);
        while (stack.length > 0) {
            const traversedTile = depthFirst ? stack.pop() : stack.shift();
            const traverseChildren = await traversalCallback(traversedTile);
            if (traverseChildren) {
                const children = await traversedTile.getChildren();
                const length = children.length;
                for (let i = 0; i < length; i++) {
                    const traversedChild = children[i];
                    stack.push(traversedChild);
                }
            }
        }
    }
}
exports.TilesetTraverser = TilesetTraverser;
