/// <reference types="node" />
import { Validator } from "./Validator";
import { ContentData } from "./ContentData";
/**
 * A class for managing `Validator` instances that are used for
 * validating the data that is pointed to by a `content.uri`.
 *
 * The only public methods (for now) are `registerDefaults`,
 * which registers all known content data validators, and
 * `findContentDataValidator`, which returns the validator
 * that should be used for a given `ContentData` object.
 *
 * @internal
 */
export declare class ContentDataValidators {
    /**
     * The list of validators that have been registered.
     */
    private static readonly dataValidators;
    /**
     * Registers all default content data validators
     */
    static registerDefaults(): void;
    private static registerArchive;
    /**
     * Tries to find a data validator that can be used for validating
     * the given content data. If no matching validator can be found,
     * then `undefined` is returned.
     *
     * @param contentData - The `ContentData`
     * @returns The validator, or `undefined`
     */
    static findContentDataValidator(contentData: ContentData): Promise<Validator<ContentData> | undefined>;
    /**
     * Register a validator that should be used when the content
     * data starts with the given magic string.
     *
     * (This string is currently assumed to have length 4, but
     * this may have to be generalized in the future)
     *
     * @param magic - The magic string
     * @param bufferValidator - The validator for the buffer data
     */
    private static registerByMagic;
    /**
     * Register a validator that should be used when the content URI
     * has the given file extension
     *
     * The file extension should include the `"."` dot, and the
     * check for the file extension will be case INsensitive.
     *
     * @param extension - The extension
     * @param bufferValidator - The validator for the buffer data
     */
    private static registerByExtension;
    /**
     * Register the data validator for (external) tileset files.
     *
     * The condition of whether this validator is used for
     * given content data is that it `isProbablyTileset`.
     */
    private static registerTileset;
    /**
     * Register the data validator for glTF files.
     *
     * This refers to JSON files (not GLB files), and checks
     * whether the object that is parsed from the JSON data
     * is probably a glTF asset, as of `isProbablyGltf`.
     */
    private static registerGltf;
    /**
     * Returns whether the given content data is probably a tileset.
     *
     * The exact conditions for this method returning `true` are
     * intentionally not specified.
     *
     * @param contentData - The content data
     * @returns Whether the content data is probably a tileset
     */
    static isProbablyTileset(contentData: ContentData): Promise<boolean>;
    /**
     * Returns whether the given content data is probably a glTF
     * (not a GLB, but a glTF JSON).
     *
     * The exact conditions for this method returning `true` are
     * intentionally not specified.
     *
     * @param contentData - The content data
     * @returns Whether the content data is probably glTF
     */
    static isProbablyGltf(contentData: ContentData): Promise<boolean>;
    /**
     * Wraps the given validator for `Buffer` objects into one that
     * can be applied to `ContentData` objects, and just applies
     * the given validator to the buffer that is returned by
     * `ContentData#getData`.
     *
     * @param bufferValidator - The validator for `Buffer` objects
     * @returns The validator for `ContentData` objects
     */
    static wrapBufferValidator(bufferValidator: Validator<Buffer>): Validator<ContentData>;
    /**
     * Registers a data validator that will be used when a
     * `ContentData` matches the given predicate.
     *
     * @param predicate - The predicate
     * @param dataValidator - The data validator
     */
    private static registerByPredicate;
}
//# sourceMappingURL=ContentDataValidators.d.ts.map