"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ContentDataValidators = void 0;
const defined_1 = require("../base/defined");
const Validators_1 = require("./Validators");
const B3dmValidator_1 = require("../tileFormats/B3dmValidator");
const I3dmValidator_1 = require("../tileFormats/I3dmValidator");
const PntsValidator_1 = require("../tileFormats/PntsValidator");
const CmptValidator_1 = require("../tileFormats/CmptValidator");
const GltfValidator_1 = require("../tileFormats/GltfValidator");
const TilesetArchiveValidator_1 = require("./TilesetArchiveValidator");
const IoValidationIssue_1 = require("../issues/IoValidationIssue");
const ResourceTypes_1 = require("../io/ResourceTypes");
/**
 * A class for managing `Validator` instances that are used for
 * validating the data that is pointed to by a `content.uri`.
 *
 * The only public methods (for now) are `registerDefaults`,
 * which registers all known content data validators, and
 * `findContentDataValidator`, which returns the validator
 * that should be used for a given `ContentData` object.
 *
 * @internal
 */
class ContentDataValidators {
    /**
     * Registers all default content data validators
     */
    static registerDefaults() {
        // The validators will be checked in the order in which they are
        // registered. In the future, there might be a mechanism for
        // 'overriding' a previously registered validator.
        ContentDataValidators.registerByMagic("glTF", new GltfValidator_1.GltfValidator());
        ContentDataValidators.registerByMagic("b3dm", new B3dmValidator_1.B3dmValidator());
        ContentDataValidators.registerByMagic("i3dm", new I3dmValidator_1.I3dmValidator());
        ContentDataValidators.registerByMagic("cmpt", new CmptValidator_1.CmptValidator());
        ContentDataValidators.registerByMagic("pnts", new PntsValidator_1.PntsValidator());
        // Certain content types are known to be encountered,
        // but are not (yet) validated. These can either be
        // ignored, or cause a warning. In the future, this
        // should be configurable, probably even on a per-type
        // basis, via the command line or a config file
        const ignoreUnhandledContentTypes = false;
        let geomValidator = Validators_1.Validators.createEmptyValidator();
        let vctrValidator = Validators_1.Validators.createEmptyValidator();
        let geojsonValidator = Validators_1.Validators.createEmptyValidator();
        if (!ignoreUnhandledContentTypes) {
            geomValidator = Validators_1.Validators.createContentValidationWarning("Skipping 'geom' validation");
            vctrValidator = Validators_1.Validators.createContentValidationWarning("Skipping 'vctr' validation");
            geojsonValidator = Validators_1.Validators.createContentValidationWarning("Skipping 'geojson' validation");
        }
        ContentDataValidators.registerByMagic("geom", geomValidator);
        ContentDataValidators.registerByMagic("vctr", vctrValidator);
        ContentDataValidators.registerByExtension(".geojson", geojsonValidator);
        ContentDataValidators.registerArchive();
        ContentDataValidators.registerTileset();
        ContentDataValidators.registerGltf();
    }
    static registerArchive() {
        const predicate = async (contentData) => contentData.extension === ".3tz";
        const archiveValidator = new TilesetArchiveValidator_1.TilesetArchiveValidator();
        const validator = {
            async validateObject(inputPath, input, context) {
                const resourceResolver = context.getResourceResolver();
                const resolvedUri = resourceResolver.resolveUri(input.uri);
                const result = await archiveValidator.validateObject(inputPath, resolvedUri, context);
                return result;
            },
        };
        ContentDataValidators.registerByPredicate(predicate, validator);
    }
    /**
     * Tries to find a data validator that can be used for validating
     * the given content data. If no matching validator can be found,
     * then `undefined` is returned.
     *
     * @param contentData - The `ContentData`
     * @returns The validator, or `undefined`
     */
    static async findContentDataValidator(contentData) {
        for (const entry of ContentDataValidators.dataValidators) {
            if (await entry.predicate(contentData)) {
                return entry.dataValidator;
            }
        }
        return undefined;
    }
    /**
     * Register a validator that should be used when the content
     * data starts with the given magic string.
     *
     * (This string is currently assumed to have length 4, but
     * this may have to be generalized in the future)
     *
     * @param magic - The magic string
     * @param bufferValidator - The validator for the buffer data
     */
    static registerByMagic(magic, bufferValidator) {
        ContentDataValidators.registerByPredicate(async (contentData) => (await contentData.getMagic()) === magic, ContentDataValidators.wrapBufferValidator(bufferValidator));
    }
    /**
     * Register a validator that should be used when the content URI
     * has the given file extension
     *
     * The file extension should include the `"."` dot, and the
     * check for the file extension will be case INsensitive.
     *
     * @param extension - The extension
     * @param bufferValidator - The validator for the buffer data
     */
    static registerByExtension(extension, bufferValidator) {
        ContentDataValidators.registerByPredicate(async (contentData) => contentData.extension === extension.toLowerCase(), ContentDataValidators.wrapBufferValidator(bufferValidator));
    }
    /**
     * Register the data validator for (external) tileset files.
     *
     * The condition of whether this validator is used for
     * given content data is that it `isProbablyTileset`.
     */
    static registerTileset() {
        const predicate = async (contentData) => await ContentDataValidators.isProbablyTileset(contentData);
        const externalValidator = Validators_1.Validators.createDefaultTilesetValidator();
        const bufferValidator = Validators_1.Validators.parseFromBuffer(externalValidator);
        const contentDataValidator = ContentDataValidators.wrapBufferValidator(bufferValidator);
        ContentDataValidators.registerByPredicate(predicate, contentDataValidator);
    }
    /**
     * Register the data validator for glTF files.
     *
     * This refers to JSON files (not GLB files), and checks
     * whether the object that is parsed from the JSON data
     * is probably a glTF asset, as of `isProbablyGltf`.
     */
    static registerGltf() {
        const predicate = async (contentData) => await ContentDataValidators.isProbablyGltf(contentData);
        const bufferValidator = new GltfValidator_1.GltfValidator();
        const contentDataValidator = ContentDataValidators.wrapBufferValidator(bufferValidator);
        ContentDataValidators.registerByPredicate(predicate, contentDataValidator);
    }
    /**
     * Returns whether the given content data is probably a tileset.
     *
     * The exact conditions for this method returning `true` are
     * intentionally not specified.
     *
     * @param contentData - The content data
     * @returns Whether the content data is probably a tileset
     */
    static async isProbablyTileset(contentData) {
        const data = await contentData.getData();
        if (!(0, defined_1.defined)(data)) {
            return false;
        }
        if (!ResourceTypes_1.ResourceTypes.isProbablyJson(data)) {
            return false;
        }
        let parsedObject = undefined;
        try {
            parsedObject = JSON.parse(data.toString());
        }
        catch (error) {
            return false;
        }
        if (!(0, defined_1.defined)(parsedObject.asset)) {
            return false;
        }
        return (0, defined_1.defined)(parsedObject.geometricError) || (0, defined_1.defined)(parsedObject.root);
    }
    /**
     * Returns whether the given content data is probably a glTF
     * (not a GLB, but a glTF JSON).
     *
     * The exact conditions for this method returning `true` are
     * intentionally not specified.
     *
     * @param contentData - The content data
     * @returns Whether the content data is probably glTF
     */
    static async isProbablyGltf(contentData) {
        if (await ContentDataValidators.isProbablyTileset(contentData)) {
            return false;
        }
        const data = await contentData.getData();
        if (!(0, defined_1.defined)(data)) {
            return false;
        }
        if (!ResourceTypes_1.ResourceTypes.isProbablyJson(data)) {
            return false;
        }
        let parsedObject = undefined;
        try {
            parsedObject = JSON.parse(data.toString());
        }
        catch (error) {
            return false;
        }
        if (!(0, defined_1.defined)(parsedObject.asset)) {
            return false;
        }
        return true;
    }
    /**
     * Wraps the given validator for `Buffer` objects into one that
     * can be applied to `ContentData` objects, and just applies
     * the given validator to the buffer that is returned by
     * `ContentData#getData`.
     *
     * @param bufferValidator - The validator for `Buffer` objects
     * @returns The validator for `ContentData` objects
     */
    static wrapBufferValidator(bufferValidator) {
        return {
            async validateObject(inputPath, input, context) {
                const data = await input.getData();
                if (!(0, defined_1.defined)(data)) {
                    const message = `Could not resolve ${input.uri}`;
                    const issue = IoValidationIssue_1.IoValidationIssues.IO_WARNING(inputPath, message);
                    context.addIssue(issue);
                    return true;
                }
                const result = await bufferValidator.validateObject(inputPath, data, context);
                return result;
            },
        };
    }
    /**
     * Registers a data validator that will be used when a
     * `ContentData` matches the given predicate.
     *
     * @param predicate - The predicate
     * @param dataValidator - The data validator
     */
    static registerByPredicate(predicate, dataValidator) {
        const entry = {
            predicate: predicate,
            dataValidator: dataValidator,
        };
        ContentDataValidators.dataValidators.push(entry);
    }
}
exports.ContentDataValidators = ContentDataValidators;
/**
 * The list of validators that have been registered.
 */
ContentDataValidators.dataValidators = [];
