"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TilesetArchiveValidator = void 0;
const defined_1 = require("../base/defined");
const Validators_1 = require("./Validators");
const ArchiveResourceResolver_1 = require("../io/ArchiveResourceResolver");
const ContentValidationIssues_1 = require("../issues/ContentValidationIssues");
const IoValidationIssue_1 = require("../issues/IoValidationIssue");
const TilesetArchive3tz_1 = require("../archives/TilesetArchive3tz");
const ArchiveValidation3tz_1 = require("../archives/ArchiveValidation3tz");
/**
 * An implementation of a validator that validates a tileset
 * archive.
 *
 * The 'archive' here is represented as a `string`, assuming that
 * this string is a file in the local file system.
 */
class TilesetArchiveValidator {
    async validateObject(uri, resolvedUri, context) {
        //console.log("TilesetArchiveValidator resolvedUri is " + resolvedUri);
        // Create the archive from the given resolved URI (i.e.
        // the full archive file name). If the archive cannot
        // be opened, bail out with an IO_WARNING.
        let archive = undefined;
        try {
            archive = new TilesetArchive3tz_1.TilesetArchive3tz();
            archive.open(resolvedUri);
        }
        catch (error) {
            const message = `Could not create archive from ${uri}. ` +
                `Full archive URI is ${resolvedUri}.`;
            const issue = IoValidationIssue_1.IoValidationIssues.IO_WARNING(uri, message);
            context.addIssue(issue);
            return true;
        }
        // Obtain the ZIP index from the archive, to be able to apply
        // the ("legacy") 3TZ archive validation to the ZIP index
        const zipIndex = archive.getZipIndex();
        if ((0, defined_1.defined)(zipIndex)) {
            try {
                const indexValid = await ArchiveValidation3tz_1.ArchiveValidation3tz.validateIndex(zipIndex, resolvedUri, false);
                if (!indexValid) {
                    const message = `The 3TZ index is not valid`;
                    const issue = ContentValidationIssues_1.ContentValidationIssues.CONTENT_VALIDATION_ERROR(resolvedUri, message);
                    context.addIssue(issue);
                    return false;
                }
            }
            catch (error) {
                const message = `Error while validating 3TZ index: ${error}.`;
                const issue = ContentValidationIssues_1.ContentValidationIssues.CONTENT_VALIDATION_ERROR(resolvedUri, message);
                context.addIssue(issue);
                return false;
            }
        }
        // Create the `ArchiveResourceResolver` from the archive,
        // and obtain the data for the `tileset.json` file.
        // This has to be present according to the 3TZ specification.
        const archiveResourceResolver = new ArchiveResourceResolver_1.ArchiveResourceResolver("./", resolvedUri, archive);
        const tilesetJsonBuffer = await archiveResourceResolver.resolveData("tileset.json");
        if (!(0, defined_1.defined)(tilesetJsonBuffer)) {
            const message = `Could not read 'tileset.json' from archive ${uri}.` +
                `Full archive URI is ${resolvedUri}.`;
            const issue = IoValidationIssue_1.IoValidationIssues.IO_ERROR(uri, message);
            context.addIssue(issue);
            return false;
        }
        // Parse the tileset object from the JSON data
        let tileset = undefined;
        try {
            tileset = JSON.parse(tilesetJsonBuffer.toString());
        }
        catch (error) {
            const message = `Could not parse tileset JSON from 'tileset.json' ` +
                `data in archive ${uri}. Full archive URI is ${resolvedUri}.`;
            const issue = IoValidationIssue_1.IoValidationIssues.IO_ERROR(uri, message);
            context.addIssue(issue);
            return false;
        }
        // Open a new context for collecting the issues that
        // are caused by the tileset, and validate the
        // tileset using a default tileset validator.
        const derivedContext = context.deriveFromResourceResolver(archiveResourceResolver);
        const tilesetValidator = Validators_1.Validators.createDefaultTilesetValidator();
        const result = await tilesetValidator.validateObject(uri + "/tileset.json", tileset, derivedContext);
        const derivedResult = derivedContext.getResult();
        const issue = ContentValidationIssues_1.ContentValidationIssues.createForContent(uri, derivedResult);
        if (issue) {
            context.addIssue(issue);
        }
        return result;
    }
}
exports.TilesetArchiveValidator = TilesetArchiveValidator;
