"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ValidationContext = void 0;
const ValidationOptions_1 = require("./ValidationOptions");
const ValidationResult_1 = require("./ValidationResult");
/**
 * A class describing the context in which a validation takes place.
 *
 * This class is used to gather the `ValidationIssue` objects
 * that are created during the validation, and add them to a
 * `ValidationResult`.
 *
 * It also provides a `ResourceResolver` to read resources from URIs,
 * in the respective context. For example: When an external tileset
 * is encountered, a new context is derived from the given one,
 * which resolved resources against the base directory of the
 * external tileset.
 *
 * The different configuration settings that affect the behavior
 * of the validator (for example, whether external resources should
 * be validated) are stored in a `ValidationOptions` object.
 *
 */
class ValidationContext {
    constructor(resourceResolver) {
        this._options = new ValidationOptions_1.ValidationOptions();
        this._result = ValidationResult_1.ValidationResult.create();
        this._resourceResolver = resourceResolver;
        this._extensionsFound = new Set();
    }
    /**
     * Derives a new context from this one.
     *
     * It uses the same `ValidationOptions` as this one. The internal
     * `ResourceResolver` is derived by resolving the given path
     * against the original `ResourceResolver`, yielding one that
     * resolves resources against the resulting path.
     *
     * @param uri - The (usually relative) URI
     * @returns The new instance
     */
    deriveFromUri(uri) {
        const derivedResourceResolver = this._resourceResolver.derive(uri);
        return this.deriveFromResourceResolver(derivedResourceResolver);
    }
    /**
     * Derives a new context from this one.
     *
     * It uses the same `ValidationOptions` as this one, but the
     * given resource resolver.
     *
     * @param resourceResolver - The `ResourceResolver`
     * @returns The new instance
     */
    deriveFromResourceResolver(resourceResolver) {
        const derived = new ValidationContext(resourceResolver);
        derived._options = this._options;
        derived._extensionsFound = this._extensionsFound;
        return derived;
    }
    addIssue(issue) {
        this._result.add(issue);
    }
    addExtensionFound(extension) {
        this._extensionsFound.add(extension);
    }
    getExtensionsFound() {
        return new Set(this._extensionsFound);
    }
    getResult() {
        return this._result;
    }
    getResourceResolver() {
        return this._resourceResolver;
    }
    getOptions() {
        return this._options;
    }
}
exports.ValidationContext = ValidationContext;
