"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PropertyTableValidator = void 0;
const defined_1 = require("../../base/defined");
const defaultValue_1 = require("../../base/defaultValue");
const BasicValidator_1 = require("./../BasicValidator");
const RootPropertyValidator_1 = require("./../RootPropertyValidator");
const ExtendedObjectsValidators_1 = require("./../ExtendedObjectsValidators");
const MetadataStructureValidator_1 = require("./MetadataStructureValidator");
const PropertyTablePropertyValidator_1 = require("./PropertyTablePropertyValidator");
/**
 * A class for validations related to `propertyTable` objects.
 *
 * This class performs the basic JSON-level validation of the
 * property table. The validation of binary data is done with
 * the BinaryPropertyTableValidator.
 *
 * @internal
 */
class PropertyTableValidator {
    /**
     * Performs the validation to ensure that the given object is a
     * valid `propertyTable` object.
     *
     * @param path - The path for the `ValidationIssue` instances
     * @param propertyTable - The object to validate
     * @param numBufferViews - The number of buffer views that are available
     * @param schema - The `Schema` object
     * @param context - The `ValidationContext` that any issues will be added to
     * @returns Whether the object was valid
     */
    static validatePropertyTable(path, propertyTable, numBufferViews, schema, context) {
        // Make sure that the given value is an object
        if (!BasicValidator_1.BasicValidator.validateObject(path, "propertyTable", propertyTable, context)) {
            return false;
        }
        let result = true;
        // Validate the object as a RootProperty
        if (!RootPropertyValidator_1.RootPropertyValidator.validateRootProperty(path, "propertyTable", propertyTable, context)) {
            result = false;
        }
        // Perform the validation of the object in view of the
        // extensions that it may contain
        if (!ExtendedObjectsValidators_1.ExtendedObjectsValidators.validateExtendedObject(path, propertyTable, context)) {
            result = false;
        }
        // If there was an extension validator that overrides the
        // default validation, then skip the remaining validation.
        if (ExtendedObjectsValidators_1.ExtendedObjectsValidators.hasOverride(propertyTable)) {
            return result;
        }
        // Validate that the class and properties are structurally
        // valid and comply to the metadata schema
        const className = propertyTable.class;
        const tableProperties = propertyTable.properties;
        if (!MetadataStructureValidator_1.MetadataStructureValidator.validateMetadataStructure(path, "property table", className, tableProperties, schema, context)) {
            // Bail out early if the structure is not valid!
            return false;
        }
        // Validate the name.
        // If the name is defined, it MUST be a string.
        if (!BasicValidator_1.BasicValidator.validateOptionalString(path, propertyTable, "name", context)) {
            result = false;
        }
        // Validate the count
        // The count MUST be an integer
        // The count MUST be at least 1
        const count = propertyTable.count;
        const countPath = path + "/count";
        if (!BasicValidator_1.BasicValidator.validateIntegerRange(countPath, "count", count, 0, true, undefined, false, context)) {
            result = false;
        }
        // Here, the basic structure of the class and properties
        // have been determined to be valid. Continue to validate
        // the values of the properties.
        const validProperties = (0, defaultValue_1.defaultValue)(tableProperties, {});
        const validPropertyNames = Object.keys(validProperties);
        const classes = (0, defaultValue_1.defaultValue)(schema.classes, {});
        const metadataClass = classes[className];
        const classProperties = (0, defaultValue_1.defaultValue)(metadataClass.properties, {});
        // Validate each property
        for (const propertyName of validPropertyNames) {
            const propertyPath = path + "/properties/" + propertyName;
            const classProperty = classProperties[propertyName];
            // Note: The check whether 'required' properties are
            // present and have values was already done by the
            // MetadataStructureValidator
            const propertyValue = tableProperties[propertyName];
            if ((0, defined_1.defined)(propertyValue)) {
                if (!PropertyTablePropertyValidator_1.PropertyTablePropertyValidator.validatePropertyTableProperty(propertyPath, propertyName, propertyValue, numBufferViews, classProperty, context)) {
                    result = false;
                }
            }
        }
        return result;
    }
}
exports.PropertyTableValidator = PropertyTableValidator;
